@@ -87,7 +87,7 @@ def make_weights_for_balanced_classes(images, nclasses):
     return weight
 
 
-def create_splits_(log_dir, dspec):
+def should_create_splits_(log_dir, dspec):
     if dspec.get('split_dir') and _os.path.exists(dspec.get('split_dir')) and len(list(
             _os.listdir(dspec.get('split_dir')))) > 0:
         return False
@@ -100,7 +100,7 @@ def create_splits_(log_dir, dspec):
     return True
 
 
-def init_kfolds_(dspec, args):
+def default_data_splitter_(dspec, args):
     r"""
     Initialize k-folds for given dataspec.
         If: custom splits path is given it will use the splits from there
@@ -111,7 +111,7 @@ def init_kfolds_(dspec, args):
                              save_to_dir=dspec['split_dir'], shuffle_files=True, name=dspec['name'])
     else:
         if args['split_ratio'] is None or len(args['split_ratio']) == 0:
-            args['split_ratio'] = _conf.data_split_ratio
+            args['split_ratio'] = _conf.DATA_SPLIT_RATIO
         create_ratio_split(_os.listdir(dspec['data_dir']),
                            save_to_dir=dspec['split_dir'],
                            ratio=args['split_ratio'],
