@@ -2,15 +2,14 @@ r"""
 The main core of EasyTorch
 """
 
-import math as _math
 import os as _os
 from collections import OrderedDict as _ODict
 
 import torch as _torch
 
-import easytorch.config as _config
 import easytorch.data as _etdata
 import easytorch.utils as _etutils
+from easytorch.config.status import *
 from easytorch.metrics import metrics as _base_metrics
 from easytorch.utils.logger import *
 from easytorch.utils.tensorutils import initialize_weights as _init_weights
@@ -33,7 +32,7 @@ class ETTrainer:
         self.device = _ODict()
         self.optimizer = _ODict()
 
-    def init_nn(self, **kw):
+    def init_nn(self, init_weights=True, init_optimizer=True, set_device=True):
         r"""
         Call to user implementation of:
             Initialize models.
@@ -45,16 +44,14 @@ class ETTrainer:
         self._init_nn_model()
         # Print number of parameters in all models.
         if self.args['verbose']:
-            for k, m in self.nn.items():
-                if isinstance(m, _torch.nn.Module):
-                    success(f'Total params in {k}:'
-                            f' {sum(p.numel() for p in m.parameters() if p.requires_grad)}')
+            for k, m in [(_k, _m) for _k, _m in self.nn.items() if isinstance(_m, _torch.nn.Module)]:
+                success(f'Total params in {k}:' f' {sum(p.numel() for p in m.parameters() if p.requires_grad)}')
 
-        self._init_nn_weights(**kw)
-        self._init_optimizer()
-        self._set_device()
+        if init_weights: self._init_nn_weights()
+        if init_optimizer: self._init_optimizer()
+        if set_device: self._set_device()
 
-    def _init_nn_weights(self, **kw):
+    def _init_nn_weights(self):
         r"""
         By default, will initialize network with Kaimming initialization.
         If path to pretrained weights are given, it will be used instead.
@@ -66,7 +63,7 @@ class ETTrainer:
             for mk in self.nn:
                 _init_weights(self.nn[mk])
 
-    def load_checkpoint(self, full_path, src='easytorch'):
+    def load_checkpoint(self, full_path, src=MYSELF):
         r"""
         Load checkpoint from the given path:
             If it is an easytorch checkpoint, try loading all the models.
@@ -77,7 +74,7 @@ class ETTrainer:
         except:
             chk = _torch.load(full_path, map_location='cpu')
 
-        if chk.get('source', 'Unknown').lower() == src:
+        if chk.get('_its_origin_', 'Unknown').lower() == src:
             for m in chk['models']:
                 try:
                     self.nn[m].module.load_state_dict(chk['models'][m])
@@ -109,7 +106,7 @@ class ETTrainer:
         If no GPU is present, CPU is used.
         """
         self.device['gpu'] = _torch.device("cpu")
-        if _config.cuda_available and len(self.args['gpus']) >= 1:
+        if CUDA_AVAILABLE and len(self.args['gpus']) >= 1:
             self.device['gpu'] = _torch.device(f"cuda:{self.args['gpus'][0]}")
             if len(self.args['gpus']) >= 2:
                 for model_key in self.nn:
@@ -139,32 +136,8 @@ class ETTrainer:
         """
         return _base_metrics.ETAverages(num_averages=1)
 
-    def check_previous_logs(self):
-        r"""
-        Checks if there already is a previous run and prompt[Y/N] so that
-        we avoid accidentally overriding previous runs and lose temper.
-        User can supply -f True flag to override by force.
-        """
-        if self.args['force']:
-            if self.cache['debug']: warn('Forced overriding previous logs.')
-            return
-        i = 'y'
-        if self.args['phase'] == 'train':
-            train_log = f"{self.cache['log_dir']}{_sep}{self.cache['experiment_id']}_log.json"
-            if _os.path.exists(train_log):
-                i = input(f"*** {train_log} *** \n Exists. OVERRIDE [y/n]:")
-
-        if self.args['phase'] == 'test':
-            test_log = f"{self.cache['log_dir']}{_sep}{self.cache['experiment_id']}_test_scores.csv"
-            if _os.path.exists(test_log):
-                if _os.path.exists(test_log):
-                    i = input(f"*** {test_log} *** \n Exists. OVERRIDE [y/n]:")
-
-        if i.lower() == 'n':
-            raise FileExistsError(f' ##### {self.args["log_dir"]} directory is not empty. #####')
-
-    def save_checkpoint(self, full_path, src='easytorch'):
-        checkpoint = {'source': src}
+    def save_checkpoint(self, full_path, src=MYSELF):
+        checkpoint = {'_its_origin_': src}
         for k in self.nn:
             checkpoint['models'] = {}
             try:
@@ -178,7 +151,6 @@ class ETTrainer:
             except:
                 checkpoint['optimizers'][k] = self.optimizer[k].state_dict()
         _torch.save(checkpoint, full_path)
-        # self.cache['log_dir'] + _sep + file_name
 
     def init_experiment_cache(self):
         r"""
@@ -201,7 +173,8 @@ class ETTrainer:
             - so Default heade is [Loss,Precision,Recall,F1,Accuracy]
         3. Set new log_dir based on different experiment versions on each datasets as per info. received from arguments.
         """
-        pass
+        self.cache['log_header'] = 'Loss,Precision,Recall,F1,Accuracy'
+        self.cache.update(monitor_metric='f1', metric_direction='maximize')
 
     def save_if_better(self, epoch, metrics):
         r"""
@@ -216,11 +189,9 @@ class ETTrainer:
             self.save_checkpoint(self.cache['log_dir'] + _sep + self.cache['checkpoint'])
             self.cache['best_score'] = sc
             self.cache['best_epoch'] = epoch
-            if self.args['verbose']:
-                success(f"Best Model Saved!!! : {self.cache['best_score']}")
+            success(f"Best Model Saved!!! : {self.cache['best_score']}", self.args['verbose'])
         else:
-            if self.args['verbose']:
-                warn(f"Not best: {sc}, {self.cache['best_score']} in ep: {self.cache['best_epoch']}")
+            warn(f"Not best: {sc}, {self.cache['best_score']} in ep: {self.cache['best_epoch']}", self.args['verbose'])
 
     def iteration(self, batch):
         r"""
@@ -243,7 +214,7 @@ class ETTrainer:
             -we need to keep track of loss
             -we need to keep track of metrics
         """
-        return {'metrics': _base_metrics.ETMetrics(), 'averages': _base_metrics.ETAverages(num_averages=1)}
+        return {}
 
     def save_predictions(self, dataset, its):
         r"""
@@ -256,21 +227,17 @@ class ETTrainer:
         pass
 
     def evaluation(self, split_key=None, save_pred=False, dataset_list=None):
-        r"""
-        Evaluation phase that handles validation/test phase
-        split-key: the key to list of files used in this particular evaluation.
-        The program will create k-splits(json files) as per specified in --nf -num_of_folds
-         argument with keys 'train', ''validation', and 'test'.
-        """
+        info(f' {split_key} ...', self.args['verbose'])
+
         for k in self.nn:
             self.nn[k].eval()
 
-        if self.args['verbose']:
-            info(f'---Running {split_key}---')
+        eval_avg, eval_metrics = self.new_averages(), self.new_metrics()
 
-        eval_avg = self.new_averages()
-        eval_metrics = self.new_metrics()
-        loaders = [_etdata.ETDataLoader.new(mode='eval', shuffle=False, dataset=d, **self.args) for d in dataset_list]
+        loaders = []
+        for dataset in dataset_list:
+            loaders.append(_etdata.ETDataLoader.new(mode=split_key if split_key is not None else 'eval',
+                                                    shuffle=False, dataset=dataset, **self.args))
         with _torch.no_grad():
             for loader in loaders:
                 its = []
@@ -282,42 +249,53 @@ class ETTrainer:
                     if not it.get('metrics'):
                         it['metrics'] = _base_metrics.ETMetrics()
 
-                    metrics.accumulate(it['metrics'])
-                    avg.accumulate(it['averages'])
+                    if not it.get('averages'):
+                        it['averages'] = _base_metrics.ETAverages()
+
+                    metrics.accumulate(it['metrics']), avg.accumulate(it['averages'])
                     if save_pred:
                         its.append(it)
-                    if self.args['verbose'] and len(dataset_list) <= 1 and i % int(_math.log(i + 1) + 1) == 0:
-                        info(f"Itr:{i}/{len(loader)}, {it['averages'].get()}, {it['metrics'].get()}")
+
+                    if self.args['verbose'] and len(dataset_list) <= 1 and lazy_debug(i):
+                        info(f" Itr:{i}/{len(loader)}, {it['averages'].get()}, {it['metrics'].get()}")
 
                 eval_metrics.accumulate(metrics)
                 eval_avg.accumulate(avg)
                 if self.args['verbose'] and len(dataset_list) > 1:
                     info(f"{split_key}, {avg.get()}, {metrics.get()}")
                 if save_pred:
-                    self.save_predictions(loader.dataset, its)
+                    self.save_predictions(loader.dataset, self._reduce_iteration(its))
 
-        if self.args['verbose']:
-            info(f"{self.cache['experiment_id']} {split_key} metrics: {eval_metrics.get()}")
+        info(f"{self.cache['experiment_id']} {split_key} metrics: {eval_avg.get()}, {eval_metrics.get()}",
+             self.args['verbose'])
         return eval_avg, eval_metrics
 
     def _reduce_iteration(self, its):
         if len(its) == 1:
             return its[0]
         reduced = {}.fromkeys(its[0].keys(), None)
-        for k in reduced:
-            if isinstance(its[0][k], _base_metrics.ETAverages):
-                reduced[k] = self.new_averages()
-                [reduced[k].accumulate(ik[k]) for ik in its]
+        for key in reduced:
+            if isinstance(its[0][key], _base_metrics.ETAverages):
+                reduced[key] = self.new_averages()
+                [reduced[key].accumulate(ik[key]) for ik in its]
 
-            elif isinstance(its[0][k], _base_metrics.ETMetrics):
-                reduced[k] = self.new_metrics()
-                [reduced[k].accumulate(ik[k]) for ik in its]
-
-            elif isinstance(its[0][k], _torch.Tensor) and not its[0][k].requires_grad and its[0][k].is_leaf:
-                reduced[k] = _torch.cat([ik[k] for ik in its])
+            elif isinstance(its[0][key], _base_metrics.ETMetrics):
+                reduced[key] = self.new_metrics()
+                [reduced[key].accumulate(ik[key]) for ik in its]
+            elif isinstance(its[0][key], _torch.Tensor) and not its[0][key].requires_grad and its[0][key].is_leaf:
+                def collect(k=key, src=its):
+                    _data = []
+                    for ik in src:
+                        if len(ik[k].shape) > 0:
+                            _data.append(ik[k])
+                        else:
+                            _data.append(ik[k].unsqueeze(0))
+                    return _torch.cat(_data)
 
+                reduced[key] = collect
             else:
-                reduced[k] = [ik[k] for ik in its]
+                reduced[key] = (ik[key] for ik in its)
+
         return reduced
 
     def _on_epoch_end(self, **kw):
@@ -340,9 +318,9 @@ class ETTrainer:
         """
         return kw.get('epoch') - self.cache['best_epoch'] >= self.args.get('patience', 'epochs')
 
-    def _plot_progress(self, epoch, *kw):
+    def _save_progress(self, epoch):
         _log_utils.plot_progress(self.cache, experiment_id=self.cache['experiment_id'],
-                                 plot_keys=['training_log', 'validation_log'], epoch=epoch)
+                                 plot_keys=[LogKey.TRAIN_LOG, LogKey.VALIDATION_LOG], epoch=epoch)
 
     def training_iteration(self, i, batch):
         r"""
@@ -351,51 +329,59 @@ class ETTrainer:
         """
         it = self.iteration(batch)
         it['loss'].backward()
-        if i % self.args.get('num_iteration', 1) == 0:
+        if i % self.args.get('num_iterations', 1) == 0:
             first_optim = list(self.optimizer.keys())[0]
             self.optimizer[first_optim].step()
             self.optimizer[first_optim].zero_grad()
         return it
 
     def train(self, dataset, val_dataset):
+        info('Training ...', self.args['verbose'])
+
+        local_iter = self.args.get('num_iterations', 1)
         train_loader = _etdata.ETDataLoader.new(mode='train', shuffle=True, dataset=dataset, **self.args)
-        local_iter = self.args.get('num_iteration', 1)
-        total_iter = len(train_loader) // local_iter
+        tot_iter = len(train_loader) // local_iter
+
         for epoch in range(1, self.args['epochs'] + 1):
             for k in self.nn:
                 self.nn[k].train()
 
-            _metrics, _avg = self.new_metrics(), self.new_averages()
+            _metrics, _avg, its = self.new_metrics(), self.new_averages(), []
             ep_avg, ep_metrics = self.new_averages(), self.new_metrics()
-            its = []
-            it = {'averages': _base_metrics.ETAverages(), 'metrics': _base_metrics.ETMetrics()}
+
             for i, batch in enumerate(train_loader, 1):
                 its.append(self.training_iteration(i, batch))
                 if i % local_iter == 0:
                     it = self._reduce_iteration(its)
-                    _i, its = i // local_iter, []
+                    if not it.get('metrics'):
+                        it['metrics'] = _base_metrics.ETMetrics()
+                    if not it.get('averages'):
+                        it['averages'] = _base_metrics.ETAverages()
 
                     ep_avg.accumulate(it['averages']), ep_metrics.accumulate(it['metrics'])
                     _avg.accumulate(it['averages']), _metrics.accumulate(it['metrics'])
-                    if self.args['verbose'] and (lazy_debug(_i) or _i == total_iter):
-                        info(f"Ep:{epoch}/{self.args['epochs']},Itr:{_i}/{total_iter},{_avg.get()},{_metrics.get()}")
-                        self.cache['training_log'].append([*_avg.get(), *_metrics.get()])
+
+                    _i, its = i // local_iter, []
+                    if lazy_debug(_i) or _i == tot_iter:
+                        info(f"Ep:{epoch}/{self.args['epochs']},Itr:{_i}/{tot_iter},{_avg.get()},{_metrics.get()}",
+                             self.args['verbose'])
+                        self.cache[LogKey.TRAIN_LOG].append([*_avg.get(), *_metrics.get()])
                         _metrics.reset(), _avg.reset()
                     self._on_iteration_end(i=_i, epoch=epoch, it=it)
 
-            self.cache['training_log'].append([*ep_avg.get(), *ep_metrics.get()])
             val_averages, val_metric = self.evaluation(split_key='validation', dataset_list=[val_dataset])
-            self.cache['validation_log'].append([*val_averages.get(), *val_metric.get()])
+            self.cache[LogKey.VALIDATION_LOG].append([*val_averages.get(), *val_metric.get()])
             self.save_if_better(epoch, val_metric)
 
             self._on_epoch_end(epoch=epoch, epoch_averages=ep_avg, epoch_metrics=ep_metrics,
                                validation_averages=val_averages, validation_metric=val_metric)
 
             if lazy_debug(epoch):
-                self._plot_progress(epoch=epoch)
+                self._save_progress(epoch=epoch)
 
             if self._stop_early(epoch=epoch, epoch_averages=ep_avg, epoch_metrics=ep_metrics,
                                 validation_averages=val_averages, validation_metric=val_metric):
                 break
+
         """Plot at the end regardless."""
-        self._plot_progress(epoch=epoch)
+        self._save_progress(epoch=epoch)
